/*#======================================================================
# This source code is public domain, and may be freely distributed,
# modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO
# WARRANTY OF ANY KIND.
#======================================================================
# _DY_CONTRIB_PGM_:du_pipeio.c:Write to/from pipes
#======================================================================
# LICENSE MODULES: toolkits:
#                  programs:
#                  servers:
# AUTHORS: 
# DESCRIPTION:Write to/from pipes. 
#======================================================================*/

#include <stdio.h>
#include <string.h>
#include "dt_smiles.h"
#include "du_pipeio.h"

#define FALSE   0
#define TRUE    0

typedef struct {
   dt_Handle hello;
   dt_Handle (*talk)(dt_Handle sos);
} A;

static A *a = NULL;

/*============================================================================
 *  zapseq() - deallocate a sequence of objects
 */
 
static void zapseq(dt_Handle seq)
{
   dt_Handle  ob;
 
   dt_reset(seq);
   while (NULL_OB != (ob = dt_next(seq)))
      dt_dealloc(ob);
   dt_dealloc(seq);
}

/*============================================================================
 *  printsos() - print sequence of strings to fp
 */
 
static void printsos(FILE *fp, dt_Handle sos)
{
   dt_Handle  sob;
   char      *str;
   int        lens;
 
   dt_reset(sos);
   while (NULL_OB != (sob = dt_next(sos))) {
      str = dt_stringvalue(&lens, sob);
      fprintf(fp, "%.*s\n", lens, str);
   }
}
 
/*============================================================================
 *  printmsg() - print sequence of strings to stdout as a message.
 */
 
static void printmsg(dt_Handle sos)
{
   printsos(stdout, sos);
   puts(DU_PT_EOM);
   fflush(stdout);
}
 
/*============================================================================
 *  wait2die() -- called on eot: get farewell msg, read until EOF, exit.
 */
 
static void wait2die(void)
{
   dt_Handle sos;
   
   /*** prepare and send farewell message. ***/

   sos = a->talk(NULL_OB);
   printmsg(sos);
   zapseq(sos);

   /*** Read until EOF and exit. ***/

   while (EOF != getchar())
     ;
   exit(0);
}
 
/*============================================================================
 *  getmsg() -- get next message on pipe as an sos
 *
 *  This returns an sos on success; NULL_OB means EOT or error reading pipe.
 *  Should we buffer?
 */
 
static dt_Handle getmsg(void)
{
   dt_Handle sos;
   char      buf[MX_PIPEBUF];
 
   /*** Save input until DU_PT_EOM (end of msg) or DU_PT_EOT (game over). ***/
 
   sos = dt_alloc_seq();

   while (1) {
      if (NULL == fgets(buf,MX_PIPEBUF,stdin)          ) { zapseq(sos); return NULL_OB; }
      buf[strlen(buf)-1]='\0';
      if (0    == strcmp(DU_PT_EOM, buf)) break;
      if (0    == strcmp(DU_PT_EOT, buf)) { zapseq(sos); wait2die(); }
      dt_append(sos, dt_alloc_string(strlen(buf), buf));
   }

   /*** Return sos. ***/

   dt_reset(sos);
   return sos;
}

/*============================================================================
 *  du_pipe_error() - print sequence of strings to stderr
 */
 
int du_pipe_error(dt_Handle sos)
{
   printsos(stderr, sos);
   return 1;
}
 
/*============================================================================
 *  du_pipe_loop() -- initialize the pipe and start an event loop
 */

int du_pipe_loop(dt_Handle hello, dt_Handle (*talk)(dt_Handle sos))
{
   dt_Handle msgin, msgout, sos, sob;

   if (a != NULL) return 0;

   a        = (A *) malloc(sizeof(A));
   a->talk  = talk;

   /*** Send first message. ***/

   printmsg(hello);

   /*** Then loop over input, sending output. ***/

   while (NULL_OB != (msgin = getmsg())) {
      msgout = a->talk(msgin);
      printmsg(msgout);
      zapseq(msgin);
      zapseq(msgout);
   }

   /*** Never returns successfully. ***/

   exit(1);
}
